///////////////////////////////////////////////////////////////////////////////
//
// Developed by Gardener of Thoughts (gardenerofthoughts.org).
// Released under a "Common Sense License".
//
///////////////////////////////////////////////////////////////////////////////

using System;
using System.ComponentModel;
using Fluogen.Dependz;
using Fluogen.Dependz.Configuration;
using Got.Libraries;

namespace Fluogen.DeployedScripts.Configuration.FluonGenerators
{
	public class Illumination : ScriptConfiguration
	{
		public enum SurfacePerturbation
		{
			None,
			DoubleWhammyX,
			DoubleWhammyY,
			DoubleWhammyZ,
			DoubleWhammyXY,
			DoubleWhammyXZ,
			DoubleWhammyYZ,
			DoubleWhammyXYZ
		}

		public const double OrbitSeparationFractionDef = 1E-5;
		public const SurfacePerturbation PerturbationDef = SurfacePerturbation.None;
		public const TransferFunction PerturbationInnerTransferFunctionDef = TransferFunction.Invert;
		public const TransferFunction PerturbationOuterTransferFunctionDef = TransferFunction.Unchanged;
		public const double PerturbationInnerAnchorDef = 1;
		public const double HorizontalAngleOfLightDef = 45;
		public const double VerticalAngleOfLightDef = 10;

		public override string ScriptDescription => "3D illumination effect";

		[DefaultValue( 0 )]
		[AutoRenderOnConfigurationChange]
		[Description( "The number of the horizontal axis with which to create the illumination effect."
			+ " If this is 0 or a negative number, the number of the horizontal axis of the stratum is used instead." )]
		public int AxisNumberForWidth { get; set; }

		[DefaultValue( 0 )]
		[AutoRenderOnConfigurationChange]
		[Description( "The number of the vertical axis with which to create the illumination effect."
			+ " If this is 0 or a negative number, the number of the vertical axis of the stratum is used instead." )]
		public int AxisNumberForHeight { get; set; }

		[DefaultValue( OrbitSeparationFractionDef )]
		[AutoRenderOnConfigurationChange]
		[Description( "The orbit separation of each additional point used to create the illumination effect, expressed as"
			+ " a fraction from the distance between the pixels of the stratum." )]
		public double OrbitSeparationFraction { get; set; } = OrbitSeparationFractionDef;

		[DefaultValue( false )]
		[AutoRenderOnConfigurationChange]
		[Description( "If this is true, the smallest distance between pixels is used to compute the orbit separation for both the"
			+ " width and the height of the stratum." )]
		public bool UseSmallestOrbitSeparation { get; set; } = false;

		[DefaultValue( PerturbationDef )]
		[AutoRenderOnConfigurationChange]
		[Description( "The perturbation function for the surface." )]
		public SurfacePerturbation Perturbation { get; set; } = PerturbationDef;

		[DefaultValue( PerturbationInnerAnchorDef )]
		[AutoRenderOnConfigurationChange]
		[Description( "The inner anchor for the surface perturbation." )]
		public double PerturbationInnerAnchor { get; set; } = PerturbationInnerAnchorDef;

		[DefaultValue( PerturbationInnerTransferFunctionDef )]
		[AutoRenderOnConfigurationChange]
		[Description( "The inner transfer function for the surface perturbation." )]
		public TransferFunction PerturbationInnerTransferFunction { get; set; }
			= PerturbationInnerTransferFunctionDef;

		[DefaultValue( PerturbationOuterTransferFunctionDef )]
		[AutoRenderOnConfigurationChange]
		[Description( "The outer transfer function for the surface perturbation." )]
		public TransferFunction PerturbationOuterTransferFunction { get; set; }
			= PerturbationOuterTransferFunctionDef;

		[DefaultValue( HorizontalAngleOfLightDef )]
		[AutoRenderOnConfigurationChange]
		[Description( "The horizontal angle of the illumination effect. It's not an exact angle, it depends on the algorithm, but"
			+ " does control the horizontal angle." )]
		public double HorizontalAngleOfLight { get; set; } = HorizontalAngleOfLightDef;

		[DefaultValue( VerticalAngleOfLightDef )]
		[AutoRenderOnConfigurationChange]
		[Description( "The vertical angle of the illumination effect." )]
		public double VerticalAngleOfLight { get; set; } = VerticalAngleOfLightDef;

		[DefaultValue( true )]
		[AutoRenderOnConfigurationChange]
		[Description( "If this is true, the \"AmbientLight\" is used as the minimum amount of light. If this is true, the color"
			+ " fragmentation in very detailed areas is reduced, but clipping may occur. Clipping may be so severe that large parts"
			+ " of the image could be empty / missing." )]
		public bool UseMinimumAmbientLight { get; set; } = true;

		[DefaultValue( 0.0 )]
		[AutoRenderOnConfigurationChange]
		[Description( "The minimum ambient light." )]
		public double AmbientLight { get; set; }

		public override void Validate( IFluonConfiguration fc )
		{
			base.Validate( fc );

			int axisNumberForWidth = AxisNumberForWidth > 0 ? AxisNumberForWidth : fc.AxisNumberForStratumWidth;
			int axisNumberForHeight = AxisNumberForHeight > 0 ? AxisNumberForHeight : fc.AxisNumberForStratumHeight;

			fc.DomainOrigin.ValidateCartesianIndex( axisNumberForWidth - 1 );
			fc.DomainOrigin.ValidateCartesianIndex( axisNumberForHeight - 1 );

			if( axisNumberForWidth == axisNumberForHeight )
				throw new ArgumentException( $"{nameof( Illumination )}: AxisNumberForWidth == AxisNumberForHeight" );
		}
	}
}
